"""Docker commands."""
import sys
import logging
from sh import docker, ErrorReturnCode

from .base import CliCommand
from .dotci3 import DotCi3Mixin
from ci3.error import Ci3Error


logger = logging.getLogger(__name__)


class BuildCommand(CliCommand, DotCi3Mixin):
    """Build container images with docker."""

    def run(self, args):
        """Call docker to build image."""
        self.load_vars()
        for name in self.config_vars['containers']:
            values = self.config_vars['containers'][name]
            image_registry_url = self.config_vars['cluster']['image_registry_url']
            # Tag with branch name.
            tag = "{}/{}:{}".format(
                image_registry_url,
                values['image']['name'],
                self.git_branch_ending())
            try:
                logger.info('Building %s..' % name)
                options = ['.']
                if 'build' in values and 'options' in values['build']:
                    options = values['build']['options'].split(" ") + options
                docker.build('-t', tag, *options, _out=sys.stdout, _err=sys.stderr)
                logger.info('Done')
            except ErrorReturnCode as error:
                raise Ci3Error("Failed to build docker image `{}`: {}"
                               .format(name, error))


class PushCommand(CliCommand, DotCi3Mixin):
    """Push container images to docker registry."""

    def _push(self, tag):
        logger.info('Pushing %s..' % tag)
        if self.config_vars['cluster']['type'] == 'gke':
            from .gke import push_image
            push_image(tag)
        else:
            docker.push(tag, _out=sys.stdout, _err=sys.stderr)
        logger.info('Done')

    def _tag_remote(self, exising_tag, new_tag):
        if self.config_vars['cluster']['type'] == 'gke':
            from .gke import tag_container
            logger.info('Adding tag %s..' % new_tag)
            tag_container(exising_tag, new_tag)
            logger.info('Done')

    def run(self, args):
        """Call docker to push image."""
        self.load_vars()
        for name in self.config_vars['containers']:
            values = self.config_vars['containers'][name]
            image_registry_url = self.config_vars['cluster']['image_registry_url']
            # Tag with branch name.
            tag = "{}/{}:{}".format(
                image_registry_url,
                values['image']['name'],
                self.git_branch_ending())
            try:
                # Tag with git sha
                tag_sha = "{}/{}:{}".format(
                    image_registry_url,
                    values['image']['name'],
                    'commit-' + self.get_head_sha())
                docker.tag(tag, tag_sha, _out=sys.stdout, _err=sys.stderr)
                # .. and then push
                self._push(tag_sha)
                self._tag_remote(tag_sha, tag)
            except ErrorReturnCode as error:
                raise Ci3Error("Failed to push docker image `{}`: {}"
                               .format(tag, error))



"""Google Container Engine (GKE) cli command."""
import sys
from sh import ErrorReturnCode, gcloud
from .base import CliCommand


def push_image(tag):
    """Push docker image via gcloud context to resolve permission issues."""
    gcloud.docker('--', 'push', tag, _out=sys.stdout, _err=sys.stderr)


def tag_container(existing_tag, new_tag):
    """Add tag to a remote container image via gcloud."""
    gcloud.container('images', 'add-tag', existing_tag, new_tag, '--quiet',
                     _out=sys.stdout, _err=sys.stderr)


class GkeCommand(CliCommand):
    """Interface Google Container Engine (GKE) to create k8s clusters."""

    def add_arguments(self, subparser):
        """
        Add cli arguments to the subparser.

        Note if you need only first level argument, then don't implement it.
        """

    def run(self, args):
        """Execute command."""
        try:
            res = gcloud('container')
        except ErrorReturnCode as error:
            print(error)



"""
kubic-ci command line interface called `kubic`.

See:
    setup.py
"""
import argparse

from ci3.commands.base import CommandLineInterface, CliCommand
from ci3.commands.dotci3 import StatusCommand, InitCommand, ShowCommand
from ci3.commands.k8s import ApplyCommand, AccessCommand, DeployCommand
from ci3.commands.dkr import BuildCommand, PushCommand
from ci3.commands.gke import GkeCommand
from ci3.version import __version__


BOX = u'\u2B1C'
PROMPT = u'{box}: kubic CI {version}'.format(box=BOX, version=__version__)


class RedoCommand(CliCommand):
    """Shorthand for `kubic build && kubic push && kubic deploy`."""

    def run(self, args):
        """Chain three commands."""
        BuildCommand().run(args)
        PushCommand().run(args)
        DeployCommand().run(args)


def main():
    """Entry point for the CLI."""
    # Print prompt at the start, always.
    parser = argparse.ArgumentParser(description=PROMPT)
    cli = CommandLineInterface(parser)

    # Add arguments shared between commands.
    cli.parser.add_argument('-v', '--verbose', dest='verbose',
                            action='count', default=1)

    # Add commands with respective subcommands. See run method of each class.
    cli.add_command('status', StatusCommand)
    cli.add_command('init', InitCommand)
    cli.add_command('show', ShowCommand)
    cli.add_command('apply', ApplyCommand)
    cli.add_command('access', AccessCommand)
    cli.add_command('build', BuildCommand)
    cli.add_command('push', PushCommand)
    cli.add_command('deploy', DeployCommand)
    cli.add_command('redo', RedoCommand)
    # TODO: implement
    # cli.add_command('gke', GkeCommand)

    # Parse cli arguments and execute respective command to handle them.
    cli.run()



"""More advanced ci3 commands that interact with `kubectl`."""
import os
import json
import logging
from sh import kubectl
from jinja2 import Template

from ci3.error import Ci3Error
from .base import CliCommand
from .dotci3 import DotCi3Mixin, ShowCommand, CI3_CLUSTER_NAME


logger = logging.getLogger(__name__)


def access_cluster(cluster_name, cluster_namespace='default',
                   cluster_type='minikube', echo=False):
    """Access cluster by name, type."""
    if (CI3_CLUSTER_NAME not in os.environ):
        raise Ci3Error('Missing variable CI3_CLUSTER_NAME in ENV. '
                       'Have you run `kubic access <cluster_name>?`')
    if cluster_type == 'minikube':
        cluster_context = 'minikube'
    else:
        cluster_context = "$( kubectl config 'current-context' )"
    if echo:
        command = "export CI3_CLUSTER_NAME={{ cluster.name }}\n"
        if cluster_type == 'gke':
            # Also see legacy auth
            # https://stackoverflow.com/questions/49075723/what-does-unknown-user-client-mean
            command += """
export CLOUDSDK_CONTAINER_USE_CLIENT_CERTIFICATE={% if cluster.legacy_auth %}True{% else %}False{% endif %}
gcloud auth activate-service-account --key-file {{ cluster.key_path }}
gcloud container clusters get-credentials {{ cluster.name }} \
--zone {{ cluster.zone }} --project {{ cluster.project }}
""".strip() + "\n"
        elif cluster_type == 'minikube':
            command += "eval $(minikube docker-env) \n"
        command += "kubectl config set-context %s " % cluster_context
        command += "--namespace={{ cluster.namespace }} \n"
        return command.strip()
    else:
        kubectl.config('set-context', cluster_context, '--namespace=%s' % cluster_namespace)


class ApplyCommand(ShowCommand):
    """
    Render and apply k8s configuration from the jinja2 template.

    Use current kubectl context. Make sure to run `source <(kubic access <clustername>)>`.
    See also `ci3.dotci3.ShowCommand`.
    """

    def run(self, args):
        """
        Apply k8s configuration from the jinja2 template.

        Rednder template with substituted ci3 vars. Pass k8s configuration to `kubectl`.
        """
        self.load_vars()
        k8s_config = self.render(args.tpl_path)
        kubectl.apply('-f', '-', _in=k8s_config)


class AccessCommand(CliCommand, DotCi3Mixin):
    """
    Output shell code to switch between different clusters.

    Access switch needs to modify shell ENV. This requires executing
    `source <(kubic access <clustername>)>`
    """

    def add_arguments(self, subparser):
        """Add cli arguments to command subparser."""
        subparser.add_argument('cluster_name', help="Name of the cluster")
        subparser.add_argument('-n', '--namespace', default='default',
                               help="Cluster namespace")

    def run(self, args):
        """
        Apply k8s configuration from the jinja2 template.

        Render template with substituted ci3 vars. Pass k8s configuration to `kubectl`.
        """
        self.load_vars(args.cluster_name)
        os.environ[CI3_CLUSTER_NAME] = args.cluster_name
        template = Template(access_cluster(
            cluster_name=args.cluster_name,
            cluster_namespace=args.namespace,
            cluster_type=self.config_vars['cluster']['type'],
            echo=True))
        print(template.render(self.config_vars))


class DeployCommand(CliCommand, DotCi3Mixin):
    """Deploy CI cycle by applying changed configuration to k8s cluster."""

    def add_arguments(self, subparser):
        """Add cli arguments to command subparser."""
        subparser.add_argument('-d', '--deployment',
                               help="Name of k8s deployment to patch with built container tag.")

    def _patch_deployment(self, name):
        """Get tag id of last container build and patch deployment."""
        values = self.config_vars['containers'][name]
        image_registry_url = self.config_vars['cluster']['image_registry_url']
        # We expect that actually this tag with git sha is already pushed to remote repository.
        tag_sha = "{}/{}:{}".format(
            image_registry_url,
            values['image']['name'],
            'commit-' + self.get_head_sha())
        # Actual patching
        payload = {
            "spec": {
                "template": {
                    "spec": {
                        "containers": [{
                            "name": name,
                            "image": tag_sha,
                        }],
                    },
                },
            },
        }
        kubectl.patch('deployment', name, '-p', json.dumps(payload))

    def run(self, args):
        """
        Apply k8s configuration from the jinja2 template.

        Shorthand to `kubic apply .ci3/deploy.yaml`
        """
        self.load_vars()
        k8s_config = self.render(".ci3/deploy.yaml")
        kubectl.apply('-f', '-', _in=k8s_config)

        # Always patch deployment
        if not args.deployment:
            logging.info('Not patching any containers')
        else:
            if args.deployment not in self.config_vars['containers']:
                raise Ci3Error("Container not found: %s", args.deployment)
            self._patch_deployment(args.deployment)



import logging


class LogConfigurator(object):
    """Console logging that can be configured by verbosity levels."""

    def __init__(self, root=None, root_level=logging.INFO):
        self.root = logging.getLogger() if root is None else root
        self.root.setLevel(root_level)
        self.__file_handler = None
        self.__console_handler = None

    @property
    def console_handler(self):
        if self.__console_handler is None:
            self.__console_handler = logging.StreamHandler()
        return self.__console_handler

    def map_verbosity_to_level(self, value):
        """Verbosity value is just an integer count of v-char in `-vvvv`."""
        return logging.CRITICAL - (value * 10) % logging.CRITICAL

    def set_console_handler(self, verbosity):
        self.console_handler.setLevel(self.map_verbosity_to_level(verbosity))
        # if self.root.level < self.console_handler.level:
        #     self.root.level = self.console_handler.level
        self.root.level = self.console_handler.level
        format_str = '%(asctime)s %(name)-30s %(levelname)-8s %(message)s'
        datefmt_str = '%m-%d %H:%M:%S'
        self.console_handler.setFormatter(
            logging.Formatter(format_str, datefmt_str))
        self.root.addHandler(self.console_handler)



"""Basic Cli commands to interact with the content of `.ci3` folder."""
import os
import re
import yaml
import logging
import jinja2

from ci3.error import Ci3Error
from .base import CliCommand


CI3_CLUSTER_NAME = 'CI3_CLUSTER_NAME'
logger = logging.getLogger(__name__)


class MissingDotCi3Folder(Ci3Error):
    """Raised if project not initialized and `.ci3` is missing."""


class DotCi3Mixin(object):
    """Help with `.ci3` folder project configuration."""

    @property
    def dotci3_path(self):
        """Return fullpath string to `.ci3` folder."""
        return os.path.join(os.getcwd(), '.ci3')

    def get_dotci3_path(self):
        """Return fullpath to `.ci3` if exists or raise error."""
        path = os.path.join(os.getcwd(), '.ci3')
        if not os.path.exists(path):
            raise MissingDotCi3Folder("Missing `.ci3` folder. Have you tried "
                                      "`kubic init` ?")
        return path

    @property
    def secrets_path(self):
        """Return fullpath string to `.ci3/secrets` folder."""
        return os.path.join(self.dotci3_path, 'secrets')

    @property
    def vars_path(self):
        """Return fullpath string to `.ci3/vars` folder."""
        return os.path.join(self.dotci3_path, 'vars')

    @property
    def cluster_vars_path(self):
        """Return fullpath string to `.ci3/vars/clusters` folder."""
        return os.path.join(self.vars_path, 'clusters')

    @property
    def services_path(self):
        """Return fullpath string to `.ci3/services` folder."""
        return os.path.join(self.dotci3_path, 'services')

    @property
    def deploy_path(self):
        """Return fullpath string to `.ci3/deploy.yaml` jinja2 template."""
        return os.path.join(self.dotci3_path, 'deploy.yaml')

    @staticmethod
    def git_branch_ending():
        """
        Lookup and return currently checkout git branch.

        Pick the postfix ending matching [alphanumerical, "_", "-"]. Function strictly cuts any
        prefix that does not match those criteria. E.g. "feature/foo-bar" -> "-feat"
        """
        if 'CI_COMMIT_REF_NAME' in os.environ:
            # We are inside gitlab-runner, so branches are not checkout.
            # Solution is to pick the name for them ENV variable.
            name = os.environ['CI_COMMIT_REF_NAME'].strip()
        else:
            from sh import git, ErrorReturnCode
            try:
                result = git('rev-parse', '--abbrev-ref', 'HEAD')
            except ErrorReturnCode as error:
                raise Ci3Error("Failed to get the name of the current git branch: %s" % error)
            name = result.strip()
        # Cutting the non-matching prefix.
        ending = re.search('[a-zA-Z0-9_\-]*$', name)
        if ending is None:
            raise Ci3Error("Name of the branch can be only of alphanumerical, "
                           "underscore, minus")
        return ending.group()

    @staticmethod
    def get_head_sha():
        """Get SHA1 of the local git HEAD."""
        from sh import git, ErrorReturnCode
        try:
            result = git('rev-parse', 'HEAD')
        except ErrorReturnCode as error:
            raise Ci3Error("Failed to get SHA1 of the local git HEAD: %s" % error)
        return result.strip()

    def _load_global_vars(self):
        """Load global vars from `.ci3` project folder."""
        global_vars_path = os.path.join(self.vars_path, 'global.yaml')
        if not os.path.exists(global_vars_path):
            raise IOError('Path does not exists: %s' % global_vars_path)
        with open(global_vars_path) as vars_stream:
            self.config_vars = yaml.load(vars_stream)
        if self.config_vars is None:
            self.config_vars = {}

    def _load_cluster_vars(self, cluster_name):
        """Load cluster specific vars, overwrite global values."""
        if not ('cluster' in self.config_vars and type(self.config_vars['cluster']) == dict):
            self.config_vars['cluster'] = dict()
        # Set cluster namespace
        try:
            self.config_vars['cluster']['namespace'] = self.git_branch_ending()
        except Ci3Error as error:
            logger.error(error)
            self.config_vars['cluster']['namespace'] = 'default'
        # Set cluster name
        self.config_vars['cluster']['name'] = cluster_name
        cluster_vars = self.config_vars['cluster']
        # Finally load vars and update config.
        with open(os.path.join(self.cluster_vars_path,
                               self.config_vars['cluster']['name']) + '.yaml') as vars_stream:
            self.config_vars.update(yaml.load(vars_stream))
        self.config_vars['cluster'].update(cluster_vars)

    def load_vars(self, cluster_name=None):
        """Load vars from `.ci3` project folder."""
        if not cluster_name:
            if (CI3_CLUSTER_NAME not in os.environ):
                raise Ci3Error('Missing variable CI3_CLUSTER_NAME in ENV. '
                               'Have you run `kubic access <cluster_name>?`')
            cluster_name = os.environ[CI3_CLUSTER_NAME]
        self._load_global_vars()
        self._load_cluster_vars(cluster_name)
        # Add ENV vars
        self.config_vars['env'] = os.environ
        # Add git branch.
        if 'git' not in self.config_vars:
            self.config_vars['git'] = dict()
        self.config_vars['git'].update({'branch': self.git_branch_ending()})

    def render(self, template_path, template_vars=None):
        """Render jinja2 template, apply template_vars (optional) or `self.config_vars`."""
        logger.debug('Path to k8s templates: %s' % self.dotci3_path)
        env = jinja2.Environment()
        env.loader = jinja2.FileSystemLoader(self.dotci3_path)
        template_str = open(template_path).read()
        if not template_vars:
            template_vars = self.config_vars
        return env.from_string(template_str).render(template_vars)


class StatusCommand(CliCommand, DotCi3Mixin):
    """Report status of ci3 project."""

    def run(self, args):
        """
        Report status of the project based on the content of `.ci3` folder.

        If not folder is found advice to call `init` command.
        """
        print('kubic-ci project ({})'.format(self.get_dotci3_path()))


class InitCommand(CliCommand, DotCi3Mixin):
    """Report status of ci3 project."""

    def run(self, args):
        """Populate some default structure into `.ci3` folder."""
        if os.path.exists(self.dotci3_path):
            print('kubic-ci project has been already initialized..Skipping')
            return
        print('Initializing kubic-ci project configuration: {}'
              .format(self.dotci3_path))
        # Init secrets folder
        os.makedirs(self.secrets_path)
        with open(os.path.join(self.secrets_path, '.gitignore'), 'w+') as gi:
            gi.write('# Place your keys and secrets here, but never commit')
        # Init vars: global and cluster specific
        os.makedirs(self.vars_path)
        with open(os.path.join(self.vars_path, 'global.yaml'), 'w+') as global_vars:
            global_vars.write("""
# Place your global jinja2 vars here, applicable across all clusters.
# Note that this file will be overwritten by `.ci3/vars/clusters/<cluster.name>.yaml`
---
containers:
  homepage:
    build:
      dockerfile: Dockerfile
    image:
      name: homepage
      tag: last
""".strip())
        # `.ci3/vars/clusters` and `.ci3/vars/clusters/minikube.yaml`
        os.makedirs(self.cluster_vars_path)
        with open(os.path.join(self.cluster_vars_path, 'minikube.yaml'), 'w+') as minikube_vars:
            minikube_vars.write("""
# Place your local (minikube) jinja2 vars here
---
cluster:
  type: minikube
""".strip())
        # `.ci3/namespace.yaml`
        with open(os.path.join(self.dotci3_path, 'namespace.yaml'), 'w+') as namespace_yaml:
            namespace_yaml.write("""
---
kind: Namespace
apiVersion: v1
metadata:
  name: {{ cluster.namespace }}
  labels:
    name: {{ cluster.namespace }}
""".strip())
        # `.ci3/deploy.yaml`
        os.makedirs(self.services_path)
        deploy_yaml_header = """
# Note: this template is used in `kubic show .ci3/deploy.yaml | kubectl apply -f -`, i.e.
# each time when deploy is triggered to update the k8s state from the source code. One should not
# include here k8s configuration that is required only once, during cluster construction.
{% include 'namespace.yaml' %}
{% include 'services/web_service.yaml' %}
"""
        with open(self.deploy_path, 'w+') as deploy_yaml:
            deploy_yaml.write(deploy_yaml_header.strip())

        web_service_yaml = """
# This is an example web-service.
---
kind: Deployment
apiVersion: extensions/v1beta1
metadata:
  name: homepage
  namespace: {{ cluster.namespace }}
spec:
  replicas: 1
  template:
    metadata:
      labels:
        app: homepage
    spec:
      {% if cluster.type == 'minikube' %}
      volumes:
        - name: code-volume
          hostPath:
            path: '{{ env.PWD }}'
      {% endif %}
      containers:
        - name: homepage
          image: {{ cluster.image_registry_url }}/{{ containers.homepage.image.name }}:{{ containers.homepage.image.tag }}
          imagePullPolicy: Always
          ports:
            - containerPort: 80
          {% if cluster.type == 'minikube' %}
          volumeMounts:
            - mountPath: '/opt/homepage'
              name: code-volume
          {% endif %}
      restartPolicy: Always
---
kind: Service
apiVersion: v1
metadata:
  name: homepage
  namespace: {{ cluster.namespace }}
  labels:
    app: homepage
spec:
  {% if cluster.type == 'minikube' %}
  type: NodePort
  {% else %}
  type: LoadBalancer
  loadBalancerIP: {{ cluster.external_ips.homepage }}
  {% endif %}
  ports:
  - name: http
    port: 80
    targetPort: 80
    protocol: TCP
  selector:
    app: homepage
"""
        with open(os.path.join(self.services_path, 'web_service.yaml'), 'w+') as ws:
            ws.write(web_service_yaml.strip())


class ShowCommand(CliCommand, DotCi3Mixin):
    """Render and show k8s configuration from the jinja2 template."""

    def add_arguments(self, subparser):
        """Add cli arguments to command subparser."""
        subparser.add_argument('tpl_path', help="Path to jinja2 template with k8s configuration.")

    def run(self, args):
        """
        Render and show k8s configuration from the jinja2 template.

        Substitute ci3 vars to do rendering.
        """
        self.load_vars()
        logger.debug('Config vars: %s' % self.config_vars)
        print(self.render(args.tpl_path))



"""Base classes for cli."""
import sys

from ci3.error import Ci3Error
from ci3.log import LogConfigurator


class CliCommand(object):
    """Base class for cli commands."""

    def add_arguments(self, subparser):
        """
        (Optional) Add cli arguments to the subparser.

        Note if you need only first level argument, then don't implement this.
        """

    def run(self, args):
        """Do actual work when running the command."""
        raise NotImplementedError()


class CommandLineInterface(object):
    """CLI abstraction."""

    def __init__(self, parser):
        """Initialize CLI class."""
        self.parser = parser
        self.subparsers = self.parser.add_subparsers()

    def add_command(self, name, cls):
        """Add command to handle subparsed results of the cli."""
        # Instantiate passed command class. It will get the parsed arguments
        # passed in case the command is invoked from the cli.
        command = cls()
        subparser = self.subparsers.add_parser(name)
        command.add_arguments(subparser)
        subparser.set_defaults(func=command.run)

    def run(self):
        """Run respective command to handle parsed arguments."""
        args = self.parser.parse_args()
        try:
            log = LogConfigurator()
            if 'access' in sys.argv:
                log.set_console_handler(0)
            else:
                log.set_console_handler(args.verbose)
            if not getattr(args, 'func', None):
                raise Ci3Error("Unknown command. See kubic -h for help")
            args.func(args)
        except Ci3Error as error:
            # Report ci3 errors rather as a message, not stack trace.
            print(error)
            exit(1)
